import tensorflow as tf
from tensorflow.keras.layers import Input
from tensorflow.keras.models import Model

from cvdm.architectures.unet import UNet


def mean_model(input_shape_condition: tf.TensorShape, out_channels: int) -> Model:
    mean_model_input = Input(input_shape_condition)
    unet_output = UNet(
        input_shape_condition, inputs=mean_model_input, out_filters=out_channels
    )
    mean_condition, condition_latents = unet_output
    mean_model_out: Model = Model(mean_model_input, [mean_condition, condition_latents])
    return mean_model_out



import os
from typing import Dict, Optional, Tuple, Union

import numpy as np
from matplotlib import pyplot as plt
from neptune import Run
from neptune.types import File
from skimage.util import montage
from tensorflow.keras.models import Model
from tqdm import tqdm

from cvdm.utils.metrics_utils import calculate_metrics


def ddpm_obtain_sr_img(
    x: np.ndarray,
    timesteps_test: int,
    noise_model: Model,
    schedule_model: Model,
    mu_model: Optional[Model],
    out_shape: Optional[Tuple[int, ...]] = None,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    if out_shape == None:
        out_shape = x.shape
    assert out_shape is not None
    pred_sr = np.random.normal(0, 1, out_shape)
    if mu_model is not None:
        mu_pred = mu_model.predict(x, verbose=0)[0]
    alpha_vec = np.zeros(out_shape + (timesteps_test,))
    for t in tqdm(range(timesteps_test)):
        t_inp = np.clip(
            np.ones(out_shape) * np.reshape(t / timesteps_test, (1, 1, 1, 1)),
            0,
            0.99999,
        )
        sch_params_t = schedule_model.predict([x, t_inp], verbose=0)
        alpha_t = np.clip(1 - sch_params_t[1] / timesteps_test, 1e-6, 0.99999)
        alpha_vec[..., t] = alpha_t
    gamma_vec = np.cumprod(alpha_vec, axis=-1)
    gamma_vec = np.clip(gamma_vec, 1e-10, 0.99999)
    count = 0
    pred_noise = 0
    for t in tqdm(range(timesteps_test, 1, -1)):
        z: Union[float, np.ndarray] = np.random.normal(0, 1, out_shape)
        if t == 1:
            z = 0
        alpha_t = alpha_vec[..., t - 1]
        beta_t = 1 - alpha_t
        gamma_t = gamma_vec[..., t - 1]
        gamma_tm1 = gamma_vec[..., t - 2]
        beta_factor = (1 - gamma_tm1) * beta_t / (1 - gamma_t)
        if count > 0:
            pred_sr = (
                np.sqrt(gamma_t) * pred_sr
                + np.sqrt(1 - gamma_t - beta_factor) * pred_noise
                + np.sqrt(beta_factor) * z
            )
        if mu_model is not None:
            pred_noise = noise_model.predict([pred_sr, x, mu_pred, gamma_t], verbose=0)
        else:
            pred_noise = noise_model.predict([pred_sr, x, gamma_t], verbose=0)
        pred_sr = (pred_sr - np.sqrt(1 - gamma_t) * pred_noise) / np.sqrt(gamma_t)
        count += 1
    if mu_model is not None:
        sigma = 0.5
        pred_diff = sigma * pred_sr + mu_pred
    else:
        pred_diff = pred_sr
    return pred_diff, gamma_vec, alpha_vec


def create_output_montage(
    pred_y: np.ndarray,
    gamma_vec: np.ndarray,
    y: np.ndarray,
    x: Optional[np.ndarray],
) -> np.ndarray:
    if pred_y.shape[3] > 1:
        channel_axis = 3
    else:
        channel_axis = None

    if x is not None:
        concatenated_images = np.concatenate(
            (pred_y, y, x, gamma_vec),
            axis=2,
        )
    else:
        concatenated_images = np.concatenate(
            (pred_y, y, gamma_vec),
            axis=2,
        )
    image: np.ndarray = montage(
        np.squeeze(concatenated_images),
        channel_axis=channel_axis,
    )
    return image


def log_loss(run: Optional[Run], avg_loss: np.ndarray, prefix: str) -> None:
    if run is not None:
        run[f"{prefix}_loss_sum"].log(avg_loss[0])
        run[f"{prefix}_loss_delta_noise"].log(avg_loss[1])
        run[f"{prefix}_loss_beta"].log(avg_loss[2])
        run[f"{prefix}_loss_KL"].log(avg_loss[3])
        run[f"{prefix}_loss_gamma"].log(avg_loss[4])
        if len(avg_loss) == 6:
            run[f"{prefix}_loss_mean"].log(avg_loss[5])
    else:
        loss_labels = [
            "Loss Sum",
            "Delta Noise Loss",
            "Beta Loss",
            "KL Loss",
            "Gamma Loss",
        ]
        formatted_losses = [
            f"{label}: {loss:.6f}" for label, loss in zip(loss_labels, avg_loss[:5])
        ]
        for loss in formatted_losses:
            print(loss)
        if len(avg_loss) == 6:
            print(f"Mean Loss: {avg_loss[5]:.6f}")


def log_metrics(
    run: Optional[Run], metrics_dict: Dict[str, float], prefix: str
) -> None:
    if run is not None:
        for metric_name, metric_value in metrics_dict.items():
            run[f"{prefix}_" + metric_name].log(metric_value)
    else:
        print(f"{prefix.capitalize()} Metrics:")
        for metric_name, metric_value in metrics_dict.items():
            print(f"{metric_name}: {metric_value:.6f}")


def save_weights(
    run: Optional[Run],
    model: Model,
    mu_model: Optional[Model],
    step: int,
    output_path: str,
    run_id: str,
) -> None:
    weights_dir = f"{output_path}/weights"
    os.makedirs(weights_dir, exist_ok=True)

    model_weights_path = f"{weights_dir}/model_{str(step)}_{run_id}.h5"
    model.save_weights(model_weights_path, True)

    if run is not None:
        run[f"model_weights/model_{str(step)}.h5"].upload(model_weights_path)

    if mu_model is not None:
        mu_model_weights_path = f"{weights_dir}/mu_model_{str(step)}_{run_id}.h5"
        mu_model.save_weights(mu_model_weights_path, True)

        if run is not None:
            run[f"mu_model_weights/mu_model_{str(step)}.h5"].upload(
                mu_model_weights_path
            )


def save_output_montage(
    run: Optional[Run],
    output_montage: np.ndarray,
    step: int,
    output_path: str,
    run_id: str,
    prefix: str,
    cmap: Optional[str] = None,
) -> None:
    output_dir = f"{output_path}/images"
    os.makedirs(output_dir, exist_ok=True)

    image_path = f"{output_dir}/{prefix}_output_{str(step)}_{run_id}.png"
    plt.imsave(image_path, output_montage, cmap=cmap)

    if run is not None:
        run[f"{prefix}_images"].append(
            File(image_path),
            description=f"Step {step}, {prefix}",
        )


def obtain_output_montage_and_metrics(
    batch_x: np.ndarray,
    batch_y: np.ndarray,
    noise_model: Model,
    schedule_model: Model,
    mu_model: Optional[Model],
    generation_timesteps: int,
    diff_inp: bool,
    task: str,
) -> Tuple[np.ndarray, Dict]:

    pred_diff, gamma_vec, _ = ddpm_obtain_sr_img(
        batch_x,
        generation_timesteps,
        noise_model,
        schedule_model,
        mu_model,
        batch_y.shape,
    )
    if diff_inp:
        pred_y = np.clip(pred_diff + batch_x, -1, 1)
    else:
        pred_y = np.clip(pred_diff, -1, 1)

    metrics = calculate_metrics(pred_y, batch_y)
    if task in ["biosr_sr", "imagenet_sr"]:
        gamma_vec = np.clip(gamma_vec[..., generation_timesteps // 2], -1, 1)
        montage_x = batch_x
    else:
        gamma_vec = np.clip(gamma_vec[..., 0:1, generation_timesteps // 2], -1, 1)
        montage_x = None
    output_montage = create_output_montage(
        pred_y,
        gamma_vec,
        batch_y,
        montage_x,
    )
    if task in ["biosr_sr", "imagenet_sr"]:
        output_montage = (output_montage * 127.5 + 127.5).astype(np.uint8)
    return output_montage, metrics



from dataclasses import dataclass


@dataclass
class EvalConfig:
    """
    Configuration settings for the evaluation process during model training.

    Attributes:
        generation_timesteps (int): Number of timesteps to use for image generation.
        output_path (str): Directory where generated images and results will be saved.
        image_freq (float): Frequency (in terms of steps) at which images are generated for inspection.
        checkpoint_freq (int): Frequency (in steps) at which the model checkpoints are saved.
        log_freq (int): Frequency (in steps) at which loss logs are recorded.
        val_freq (int): Frequency (in steps) at which validation is performed.
        val_len (int): Number of samples used for validation.
    """

    generation_timesteps: int
    output_path: str
    image_freq: float
    checkpoint_freq: int
    log_freq: int
    val_freq: int
    val_len: int



from pathlib import Path
from typing import Any, Callable, Iterator, List, Tuple

import numpy as np
import tensorflow as tf
from tensorflow.keras.models import Model

from cvdm.configs.data_config import DataConfig
from cvdm.data.image_dir_dataloader import ImageDirDataloader
from cvdm.data.npy_dataloader import NpyDataloader
from cvdm.data.phase_2shot_dataloader import Phase2ShotDataloader
from cvdm.data.phase_polychrome_dataloader import PhasePolychromeDataloader


def prepare_dataset(
    task: str, data_config: DataConfig, training: bool
) -> Tuple[tf.data.Dataset, tf.TensorShape, tf.TensorShape]:
    dataloader: Callable[
        [], Iterator[Tuple[np.ndarray[Any, Any], np.ndarray[Any, Any]]]
    ]
    if task == "biosr_sr":
        dataloader = NpyDataloader(
            path=data_config.dataset_path,
            n_samples=data_config.n_samples,
            im_size=data_config.im_size,
        )

    elif task == "imagenet_sr":
        dataloader = ImageDirDataloader(
            paths=np.array(list(Path(data_config.dataset_path).glob("*.JPEG"))),
            n_samples=data_config.n_samples,
            im_size=data_config.im_size,
        )
        x_channels = 3
        y_channels = x_channels

    elif task == "biosr_phase":
        dataloader = PhasePolychromeDataloader(
            path=data_config.dataset_path,
            n_samples=data_config.n_samples,
            im_size=data_config.im_size,
            training=training,
        )
        x_channels = 3
        y_channels = 1
    elif task in ["imagenet_phase", "hcoco_phase"]:
        dataloader = Phase2ShotDataloader(
            paths=np.array(list(Path(data_config.dataset_path).glob("*.JPEG"))),
            n_samples=data_config.n_samples,
            im_size=data_config.im_size,
            training=training,
        )
        x_channels = 2
        y_channels = 1
    elif task == "other":
        dataloader = NpyDataloader(
            path=data_config.dataset_path,
            n_samples=data_config.n_samples,
            im_size=data_config.im_size,
        )
        x_channels, y_channels = dataloader.get_channels()
    else:
        raise NotImplementedError()

    x_shape = tf.TensorShape([data_config.im_size, data_config.im_size, x_channels])
    y_shape = tf.TensorShape([data_config.im_size, data_config.im_size, y_channels])
    dataset = tf.data.Dataset.from_generator(
        dataloader,
        output_types=(tf.float32, tf.float32),
        output_shapes=(
            x_shape,
            y_shape,
        ),
    )
    return dataset, x_shape, y_shape


def prepare_model_input(
    x: np.ndarray, y: np.ndarray, diff_inp: bool = False
) -> List[np.ndarray]:
    if diff_inp:
        dfy = y - x
    else:
        dfy = y
    s_tdx = np.random.uniform(0, 1, (x.shape[0], 1, 1, 1))
    ft = s_tdx * np.ones_like(dfy)
    return [dfy, x, ft]


def train_on_batch_cvdm(
    batch_x: np.ndarray, batch_y: np.ndarray, joint_model: Model, diff_inp: bool = False
) -> np.ndarray:
    model_input = prepare_model_input(batch_x, batch_y, diff_inp)
    loss = joint_model.train_on_batch(model_input, np.zeros_like(batch_y))
    return np.array(loss)


def taylor_expand_gamma(x, n=1):
    gamma_t_n = 0
    for i in range(n):
        gamma_t_n += tf.math.pow(x[0], i + 1)

    return gamma_t_n


def time_grad(x: List[np.ndarray]) -> float:
    gamma_t = x[0]
    timestep = x[1]

    d_gamma: float = tf.gradients(
        gamma_t, timestep, unconnected_gradients=tf.UnconnectedGradients.ZERO
    )[0]
    return d_gamma



from dataclasses import dataclass


@dataclass
class ModelConfig:
    """
    Configuration settings for the model architecture and initialization.

    Attributes:
        noise_model_type (str): The type of noise model used (unet or sr3).
        alpha (float): The scaling factor for L_gamma element of CVDM loss.
        snr_expansion_n (int): The expansion factor for the Taylor expansion of SNR.
        load_weights (str): Path to the pre-trained model weights to be loaded.
        load_mu_weights (str): Path to the pre-trained weights for the mean (mu) model, if applicable.
        zmd (bool): Indicates whether Zero-Mean Diffusion (ZMD) should be used in the model.
        inp_diff (bool): Indicates whether the model should predict just the noise to remove in each diffusion step or the image with the noise already removed.
    """

    noise_model_type: str
    alpha: float
    snr_expansion_n: int
    load_weights: str
    load_mu_weights: str
    zmd: bool
    diff_inp: bool



from typing import Dict, Optional

import numpy as np
from skimage.metrics import peak_signal_noise_ratio, structural_similarity


def nmae(y_pred: np.ndarray, y_real: np.ndarray) -> float:
    nmae: float = np.sqrt(np.sum((y_pred - y_real) ** 2)) / np.sqrt(np.sum(y_real**2))
    return nmae


def calculate_metrics(
    y_pred_batch: np.ndarray,
    y_real_batch: np.ndarray,
) -> Dict[str, float]:
    y_pred_batch = np.array(y_pred_batch)
    y_real_batch = np.array(y_real_batch)

    if y_pred_batch.shape[3] > 1:
        channel_axis = 2
    else:
        channel_axis = None

    metrics = {
        "mse": np.mean((y_pred_batch - y_real_batch) ** 2),
        "mape": np.mean(np.abs((y_real_batch - y_pred_batch) / y_real_batch + 1e-10))
        * 100,
        "nmae": np.mean(
            [
                np.mean(np.abs(y_pred - y_real)) / np.mean(np.abs(y_real) + 1e-10)
                for y_pred, y_real in zip(y_pred_batch, y_real_batch)
            ]
        ),
        "psnr": np.mean(
            [
                peak_signal_noise_ratio(
                    np.squeeze(y_pred), np.squeeze(y_real), data_range=2
                )
                for y_pred, y_real in zip(y_pred_batch, y_real_batch)
            ]
        ),
        "ssim": np.mean(
            [
                structural_similarity(
                    np.squeeze(y_pred),
                    np.squeeze(y_real),
                    data_range=2,
                    channel_axis=channel_axis,
                )
                for y_pred, y_real in zip(y_pred_batch, y_real_batch)
            ]
        ),
    }

    return metrics



from typing import Optional, Tuple, Union

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import (
    Add,
    Concatenate,
    Conv2D,
    Conv2DTranspose,
    Input,
    MaxPooling2D,
)
from tensorflow.keras.models import Model
from tensorflow_addons.layers import InstanceNormalization


def tile_gamma(x: np.ndarray) -> tf.Tensor:
    gamma = x[0]
    noisy_y = x[1]
    tile_gamma: tf.Tensor = tf.ones_like(noisy_y) * tf.expand_dims(
        tf.expand_dims(gamma, axis=-1), axis=-1
    )
    return tile_gamma


def UNet(
    shape: tf.TensorShape,
    inputs: Optional[Input] = None,
    gamma_inp: Optional[Input] = None,
    out_filters: int = 2,
    base_filters: int = 64,
    full_model=False,
) -> Union[Tuple[tf.Tensor, tf.Tensor], Model]:

    if inputs is None:
        inputs = Input(shape)

    conv1 = Conv2D(
        base_filters,
        3,
        activation="softplus",
        padding="same",
    )(inputs)

    if gamma_inp is not None:
        conv_gamma = Conv2D(
            base_filters,
            7,
            activation="softplus",
            padding="same",
        )(gamma_inp)
        conv1 = Add()([conv_gamma, conv1])

    conv1 = InstanceNormalization()(conv1)
    conv1 = Conv2D(
        base_filters,
        3,
        activation="softplus",
        padding="same",
    )(conv1)
    conv1 = InstanceNormalization()(conv1)
    pool1 = MaxPooling2D(pool_size=(2, 2))(conv1)
    conv2 = (
        Conv2D(
            base_filters * 2,
            3,
            activation="softplus",
            padding="same",
        )
    )(pool1)
    conv2 = InstanceNormalization()(conv2)
    conv2 = Conv2D(
        base_filters * 2,
        3,
        activation="softplus",
        padding="same",
    )(conv2)
    conv2 = InstanceNormalization()(conv2)
    pool2 = MaxPooling2D(pool_size=(2, 2))(conv2)
    conv3 = Conv2D(
        base_filters * 4,
        3,
        activation="softplus",
        padding="same",
    )(pool2)
    conv3 = InstanceNormalization()(conv3)
    conv3 = Conv2D(
        base_filters * 4,
        3,
        activation="softplus",
        padding="same",
    )(conv3)
    conv3 = InstanceNormalization()(conv3)
    pool3 = MaxPooling2D(pool_size=(2, 2))(conv3)
    conv4 = Conv2D(
        base_filters * 8,
        3,
        activation="softplus",
        padding="same",
    )(pool3)
    conv4 = InstanceNormalization()(conv4)
    conv4 = Conv2D(
        base_filters * 16,
        3,
        activation="softplus",
        padding="same",
    )(conv4)
    conv4 = InstanceNormalization()(conv4)
    pool4 = MaxPooling2D(pool_size=(2, 2))(conv4)

    conv5 = Conv2D(
        base_filters * 16,
        3,
        activation="softplus",
        padding="same",
    )(pool4)
    conv5 = InstanceNormalization()(conv5)
    conv5 = Conv2D(
        base_filters * 16,
        3,
        activation="softplus",
        padding="same",
    )(conv5)
    conv5 = InstanceNormalization()(conv5)

    up6 = Conv2D(base_filters * 16, 2, activation="softplus", padding="same")(
        Conv2DTranspose(
            512,
            3,
            activation="softplus",
            padding="same",
            strides=2,
        )(conv5)
    )

    merge6 = Concatenate()([conv4, up6])
    conv6 = Conv2D(
        base_filters * 8,
        3,
        activation="softplus",
        padding="same",
    )(merge6)
    conv6 = InstanceNormalization()(conv6)
    conv6 = Conv2D(
        base_filters * 8,
        3,
        activation="softplus",
        padding="same",
    )(conv6)
    conv6 = InstanceNormalization()(conv6)
    up7 = Conv2D(base_filters * 4, 2, activation="softplus", padding="same")(
        Conv2DTranspose(
            256,
            3,
            activation="softplus",
            padding="same",
            strides=2,
        )(conv6)
    )
    merge7 = Concatenate()([conv3, up7])
    conv7 = Conv2D(
        base_filters * 4,
        3,
        activation="softplus",
        padding="same",
    )(merge7)
    conv7 = InstanceNormalization()(conv7)
    conv7 = Conv2D(
        base_filters * 4,
        3,
        activation="softplus",
        padding="same",
    )(conv7)
    conv7 = InstanceNormalization()(conv7)
    up8 = Conv2D(base_filters * 2, 2, activation="softplus", padding="same")(
        Conv2DTranspose(
            128,
            3,
            activation="softplus",
            padding="same",
            strides=2,
        )(conv7)
    )
    merge8 = Concatenate()([conv2, up8])
    conv8 = Conv2D(
        base_filters * 2,
        3,
        activation="softplus",
        padding="same",
    )(merge8)
    conv8 = InstanceNormalization()(conv8)
    conv8 = Conv2D(
        base_filters * 2,
        3,
        activation="softplus",
        padding="same",
    )(conv8)
    conv8 = InstanceNormalization()(conv8)

    up9 = Conv2D(base_filters, 2, activation="softplus", padding="same")(
        Conv2DTranspose(
            64,
            3,
            activation="softplus",
            padding="same",
            strides=2,
        )(conv8)
    )

    merge9 = Concatenate()([conv1, up9])
    conv9 = Conv2D(
        base_filters,
        3,
        activation="softplus",
        padding="same",
    )(merge9)
    conv9 = InstanceNormalization()(conv9)
    conv9 = Conv2D(
        base_filters,
        3,
        activation="softplus",
        padding="same",
    )(conv9)
    conv9 = InstanceNormalization()(conv9)
    conv9 = Conv2D(
        base_filters,
        3,
        activation="softplus",
        padding="same",
    )(conv9)
    conv9 = InstanceNormalization()(conv9)
    conv10 = Conv2D(
        out_filters,
        1,
        activation="linear",
        padding="same",
    )(conv9)

    if full_model and gamma_inp is not None:
        model: Model = Model([inputs, gamma_inp], conv10)
        return model
    elif full_model:
        model = Model(inputs, conv10)
        return model
    else:
        return conv10, conv5



from typing import List

import cv2
import numpy as np
import tensorflow as tf
from PIL import Image
from sklearn.feature_extraction.image import extract_patches_2d


def read_and_patch_image_from_filename(filename: str, im_size: int) -> Image.Image:
    img_orig = cv2.imread(str(filename))
    img_reg = cv2.cvtColor(img_orig, cv2.COLOR_BGR2RGB)

    if img_reg.shape[0] < im_size or img_reg.shape[1] < im_size:
        img_patch = Image.fromarray(img_reg)
        img_patch = img_patch.resize((im_size, im_size), Image.LANCZOS)
    else:
        img_patch = extract_patches_2d(img_reg, (im_size, im_size), max_patches=1)[0]
        img_patch = Image.fromarray(img_patch)
    return img_patch


def center_crop(x: np.ndarray, crop_size: int = 2048) -> np.ndarray:
    x_center = x.shape[1] // 2
    y_center = x.shape[0] // 2

    return x[
        y_center - crop_size // 2 : y_center + crop_size // 2,
        x_center - crop_size // 2 : x_center + crop_size // 2,
    ]


def obtain_noisy_sample(x: np.ndarray) -> List[np.ndarray]:
    x_0 = x[0]
    gamma = x[1]

    noise_sample = tf.random.normal(tf.shape(x_0))

    safe_sample = tf.sqrt(gamma) * x_0 + tf.sqrt(1 - gamma) * noise_sample
    safe_mean = tf.sqrt(gamma) * x_0

    return [safe_sample, noise_sample, safe_mean, 1 - gamma]



from typing import List

import tensorflow as tf


def normal_kl(data: List[float]) -> float:
    """
    Compute the KL divergence between two gaussians.
    KL divergence between normal distributions parameterized by mean and log-variance.
    """
    mean1, logvar1, mean2, logvar2 = data
    kl_loss: float = 0.5 * (
        -1.0
        + logvar2
        - logvar1
        + tf.exp(logvar1 - logvar2)
        + ((mean1 - mean2) ** 2) * tf.exp(-logvar2)
    )
    return kl_loss


def linear_loss(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
    return y_pred



from typing import Tuple

import tensorflow as tf
from tensorflow.keras.activations import swish
from tensorflow.keras.layers import AveragePooling2D, Conv2D, Input, UpSampling2D
from tensorflow.keras.models import Model
from tensorflow_addons.layers import GroupNormalization

from cvdm.architectures.components.attention_block import attention_block
from cvdm.architectures.components.deep_residual_block import (
    deep_resblock,
    up_deep_resblock,
)
from cvdm.architectures.components.residual_block import resblock, up_resblock


def upsample(x: tf.Tensor, use_conv: bool = False) -> tf.Tensor:
    channels = x.shape[3]
    x = UpSampling2D(interpolation="nearest")(x)
    if use_conv:
        assert channels is not None
        x = Conv2D(channels, 3, padding="same")(x)
    return x


def downsample(x: tf.Tensor, use_conv: bool = False):
    channels = x.shape[3]
    if use_conv:
        assert channels is not None
        x = Conv2D(channels, 3, strides=2, padding="same")(x)
    else:
        x = AveragePooling2D(strides=2, padding="same")(x)
    return x


def sr3(
    img_shape: tf.TensorShape,
    gamma_shape: tf.TensorShape,
    channel_dim: int = 64,
    channel_ramp_multiplier: Tuple[int, ...] = (1, 2, 4, 8, 8),
    attention_resolution: Tuple[int, ...] = (8,),
    out_channels: int = 3,
    num_resblock: int = 3,
    dropout: float = 0,
    use_deep_blocks: bool = False,
    resample_with_conv: bool = False,
) -> Model:
    if use_deep_blocks:
        upblock = up_deep_resblock
        block = deep_resblock
    else:
        upblock = up_resblock
        block = resblock

    num_resolutions = len(channel_ramp_multiplier)
    combined_images = Input(img_shape)
    gamma_inp = Input(gamma_shape)
    noise_level_embedding = Conv2D(
        channel_dim, 7, activation="softplus", padding="same"
    )(gamma_inp)

    x = combined_images
    x = Conv2D(channel_dim, 3, padding="same")(x)
    skip_connections = [x]
    for i, multiplier in enumerate(channel_ramp_multiplier):
        for j in range(num_resblock):
            noise_level_embedding = Conv2D(channel_dim * multiplier, 1, padding="same")(
                noise_level_embedding
            )
            x = block(
                x, noise_level_embedding, channel_dim * multiplier, dropout=dropout
            )
            if x.shape[1] in attention_resolution:
                x = attention_block(x)
            skip_connections.append(x)
        if i != num_resolutions - 1:
            x = downsample(x, use_conv=resample_with_conv)
            noise_level_embedding = downsample(
                noise_level_embedding, use_conv=resample_with_conv
            )
            skip_connections.append(x)

    x = block(
        x=x,
        noise_embedding=noise_level_embedding,
        n_out_channels=x.shape[-1],
        dropout=dropout,
    )

    x = attention_block(x)

    x = block(
        x=x,
        noise_embedding=noise_level_embedding,
        n_out_channels=x.shape[-1],
        dropout=dropout,
    )

    for i, multiplier in reversed(list(enumerate(channel_ramp_multiplier))):
        for j in range(num_resblock + 1):

            noise_level_embedding = Conv2D(channel_dim * multiplier, 1, padding="same")(
                noise_level_embedding
            )
            x = upblock(
                x,
                skip_connections.pop(),
                noise_level_embedding,
                channel_dim * multiplier,
                dropout=dropout,
            )

            if x.shape[1] in attention_resolution:
                x = attention_block(x)
        if i != 0:
            x = upsample(x, use_conv=resample_with_conv)
            noise_level_embedding = upsample(
                noise_level_embedding, use_conv=resample_with_conv
            )

    assert len(skip_connections) == 0

    x = GroupNormalization(groups=32, axis=-1)(x)
    x = swish(x)
    outputs = Conv2D(out_channels, 3, padding="same")(x)

    model: Model = Model([combined_images, gamma_inp], outputs)
    return model



from dataclasses import dataclass


@dataclass
class DataConfig:
    """
    Configuration for the dataset.

    Attributes:
        dataset_path (str): Path to the dataset directory. For NpyDataloader it is expected that it will contain x.npy and y.npy files.
        For PhasePolychromeDataloader it should contain y.npy. For the rest of dataloaders, the directory should include .JPEG images.
        n_samples (int): Number of samples to use from the dataset.
        batch_size (int): Number of samples per batch during training.
        im_size (int): The size of the patches of images (both height and width) to use.
    """

    dataset_path: str
    n_samples: int
    batch_size: int
    im_size: int



from dataclasses import dataclass


@dataclass
class NeptuneConfig:
    """
    Configuration settings for logging and tracking experiments with Neptune. Can be omitted if Neptune should not be used.

    Attributes:
        name (str): The name of te run to be logged in Neptune.
        project (str): The Neptune project identifier where the experiment will be tracked, typically in the format 'workspace/project-name'.
    """

    name: str
    project: str



from typing import Iterator, Tuple

import numpy as np

from cvdm.utils.data_utils import center_crop


class NpyDataloader:
    def __init__(
        self,
        path: str,
        n_samples: int,
        im_size: int,
    ) -> None:
        self._x = np.load(f"{path}/x.npy", mmap_mode="r+")[:n_samples]
        self._y = np.load(f"{path}/y.npy", mmap_mode="r+")[:n_samples]
        self._im_size = im_size
        self._n_samples: int = min(n_samples, self._x.shape[0])

    def __len__(self) -> int:
        return self._n_samples

    def get_channels(self) -> Tuple[int, int]:
        return self._x.shape[-1], self._y.shape[-1]

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        x, y = self._x[idx], self._y[idx]

        x = center_crop(x, crop_size=2000)
        y = center_crop(y, crop_size=2000)
        if x.shape[0] > self._im_size or x.shape[1] > self._im_size:
            center_x = np.random.randint(
                self._im_size // 2, x.shape[1] - self._im_size // 2
            )
            center_y = np.random.randint(
                self._im_size // 2, x.shape[0] - self._im_size // 2
            )

            x = x[
                center_y - self._im_size // 2 : center_y + self._im_size // 2,
                center_x - self._im_size // 2 : center_x + self._im_size // 2,
            ]
            y = y[
                center_y - self._im_size // 2 : center_y + self._im_size // 2,
                center_x - self._im_size // 2 : center_x + self._im_size // 2,
            ]
        if len(x.shape) == 2:
            x = np.expand_dims(x, -1)
            y = np.expand_dims(y, -1)
        return x, y

    def __call__(self) -> Iterator[Tuple[np.ndarray, np.ndarray]]:
        for i in range(self.__len__()):
            yield self.__getitem__(i)



from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Dict

import numpy as np
import tensorflow as tf
from tensorflow.keras.initializers import RandomNormal
from tensorflow.keras.layers import Layer


class ConditionalInstanceNormalization(Layer):
    """
    The goal of conditional instance normalization is to make the
    model aware of the amount of noise to be removed (i.e how many
    steps in the noise diffusion process are being considered). The
    implementation was informed by the appendix in
    https://arxiv.org/pdf/1907.05600.pdf and implementation at
    https://github.com/ermongroup/ncsn/blob/master/models/cond_refinenet_dilated.py
    """

    def __init__(self, **kwargs):
        super(ConditionalInstanceNormalization, self).__init__(**kwargs)

    def build(self, input_shape: np.ndarray) -> None:
        self.batch_size = input_shape[0][0]
        self.height = input_shape[0][1]
        self.width = input_shape[0][2]
        self.n_channels = input_shape[0][3]
        self.embedding_dim = input_shape[1][1]
        self.w1 = self.add_weight(
            shape=(self.embedding_dim, self.n_channels),
            initializer=RandomNormal(mean=1.0, stddev=0.02),
            trainable=True,
            name="conditional_w1",
        )
        self.b = self.add_weight(
            shape=(self.embedding_dim, self.n_channels),
            initializer="zero",
            trainable=True,
            name="conditional_b",
        )
        self.w2 = self.add_weight(
            shape=(self.embedding_dim, self.n_channels),
            initializer=RandomNormal(mean=1.0, stddev=0.02),
            trainable=True,
            name="conditional_w2",
        )

    def call(self, inputs: Iterable[tf.Tensor]) -> tf.Tensor:
        x, noise_embedding = inputs
        feature_map_means = tf.math.reduce_mean(x, axis=(1, 2), keepdims=True)
        feature_map_std_dev = tf.math.reduce_std(x, axis=(1, 2), keepdims=True) + 1e-5
        m = tf.math.reduce_mean(feature_map_means, axis=-1, keepdims=True)
        v = tf.math.reduce_std(feature_map_means, axis=-1, keepdims=True) + 1e-5
        gamma = tf.expand_dims(
            tf.expand_dims(tf.tensordot(noise_embedding, self.w1, 1), 1), 1
        )
        beta = tf.expand_dims(
            tf.expand_dims(tf.tensordot(noise_embedding, self.b, 1), 1), 1
        )
        alpha = tf.expand_dims(
            tf.expand_dims(tf.tensordot(noise_embedding, self.w2, 1), 1), 1
        )
        instance_norm = (x - feature_map_means) / feature_map_std_dev
        x = gamma * instance_norm + beta + alpha * (feature_map_means - m) / v
        return x

    def get_config(self) -> Dict[str, Any]:
        return {}

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> ConditionalInstanceNormalization:
        return cls(**config)



import keras
import tensorflow as tf
from tensorflow.keras.layers import Concatenate, Input
from tensorflow.keras.models import Model

from cvdm.architectures.sr3 import sr3
from cvdm.architectures.unet import UNet


def noise_model(
    input_shape_noisy: tf.TensorShape, out_channels: int, model_type: str, zmd=False
) -> Model:
    noisy_inp = Input(input_shape_noisy[:-1] + (out_channels,))
    ref_frame = Input(input_shape_noisy)
    c_inp = Concatenate()([noisy_inp, ref_frame])
    gamma_inp = Input(input_shape_noisy[:-1] + (out_channels,))
    model: Model

    if zmd:
        mean_inp = Input(input_shape_noisy[:-1] + (out_channels,))
        c_inp = Concatenate()([c_inp, mean_inp])

    if model_type == "sr3":
        s_model = sr3(
            keras.backend.int_shape(c_inp)[1:],
            input_shape_noisy[:-1] + (out_channels,),
            out_channels=out_channels,
        )
        noise_out = s_model([c_inp, gamma_inp])
    else:
        noise_out, _ = UNet(
            input_shape_noisy,
            inputs=c_inp,
            gamma_inp=gamma_inp,
            out_filters=out_channels,
        )

    if zmd:
        model = Model([noisy_inp, ref_frame, mean_inp, gamma_inp], noise_out)

    else:
        model = Model([noisy_inp, ref_frame, gamma_inp], noise_out)

    return model



from __future__ import annotations

from typing import Any, Dict

import numpy as np
import tensorflow as tf
from tensorflow.keras.initializers import VarianceScaling
from tensorflow.keras.layers import Attention, Layer
from tensorflow_addons.layers import GroupNormalization


class AttentionVectorLayer(Layer):
    """
    Building the query, key or value for self-attention
    from the feature map
    """

    def __init__(self, **kwargs) -> None:
        super(AttentionVectorLayer, self).__init__(**kwargs)

    def build(self, input_shape: np.ndarray) -> None:
        self.n_channels = input_shape[-1]
        self.w = self.add_weight(
            shape=(self.n_channels, self.n_channels),
            initializer=VarianceScaling(
                scale=1.0, mode="fan_avg", distribution="uniform"
            ),
            trainable=True,
            name="attention_w",
        )
        self.b = self.add_weight(
            shape=(self.n_channels,),
            initializer="zero",
            trainable=True,
            name="attention_b",
        )

    def call(self, x: tf.Tensor) -> tf.Tensor:
        out: tf.Tensor = tf.tensordot(x, self.w, 1) + self.b
        return out

    def get_config(self) -> Dict[str, Any]:
        return {}

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> AttentionVectorLayer:
        return cls(**config)


def attention_block(x: tf.Tensor) -> tf.Tensor:
    """
    Implementing self-attention block, as mentioned in
    https://arxiv.org/pdf/1809.11096.pdf
    """

    x = GroupNormalization(groups=32, axis=-1)(x)

    q = AttentionVectorLayer()(x)
    v = AttentionVectorLayer()(x)
    k = AttentionVectorLayer()(x)

    h: tf.Tensor = Attention()([q, v, k])

    return x + h



from dataclasses import dataclass


@dataclass
class TrainingConfig:
    """
    Configuration settings for the model training process.

    Attributes:
        lr (float): Learning rate for the optimizer during training.
        epochs (int): Number of complete passes through the training dataset.
    """

    lr: float
    epochs: int



import contextlib
from typing import Any, Dict, Iterator

import tensorflow as tf
from tensorflow.keras.layers import Activation, Input
from tensorflow.keras.models import Model

from cvdm.architectures.unet import UNet
from cvdm.diffusion_models.time_model import time_model


@contextlib.contextmanager
def options(options: Dict[str, Any]) -> Iterator[None]:
    old_opts = tf.config.optimizer.get_experimental_options()
    tf.config.optimizer.set_experimental_options(options)
    try:
        yield
    finally:
        tf.config.optimizer.set_experimental_options(old_opts)


def variance_model(image_shape: tf.TensorShape, out_channels: int) -> Model:
    blurry_image = Input(image_shape)
    t_input = Input(image_shape[:-1] + (out_channels,))
    t_model = time_model(image_shape[:-1] + (out_channels,), 1024, act="linear")
    t_int_model = time_model(image_shape[:-1] + (out_channels,), 1024, act="linear")
    with options({"layout_optimizer": False}):
        t_clip = tf.clip_by_value(t_input, 0, 0.99999)
    model, _ = UNet(
        blurry_image,
        inputs=blurry_image,
        gamma_inp=blurry_image,
        out_filters=out_channels,
        base_filters=32,
    )
    model = Activation("softplus")(model)
    tau = t_model(t_clip)
    beta = model * tau
    gamma = tf.clip_by_value(tf.exp(-model * t_int_model(t_clip)), 1e-6, 0.99999)
    modified_model: Model = Model(inputs=[blurry_image, t_input], outputs=[gamma, beta])

    return modified_model



from typing import Optional, Sequence, Union

import cupy as cp
import numpy as np
from cupy.fft import fftn, fftshift, ifftn, ifftshift


# If you encounter issues with cupy installation, you can use the numpy implementation of fft
def easy_fft(data: np.ndarray, axes: Optional[Sequence[int]] = None) -> cp.ndarray:
    """FFT that includes shifting."""
    return fftshift(fftn(ifftshift(data, axes=axes), axes=axes), axes=axes)


def easy_ifft(data: np.ndarray, axes: Optional[Sequence[int]] = None) -> cp.ndarray:
    """Inverse FFT that includes shifting."""
    return ifftshift(ifftn(fftshift(data, axes=axes), axes=axes), axes=axes)


def light_component_sim(
    img: np.ndarray, ps: float, z: float, central_frec: float, sigma: float
) -> np.ndarray:
    wl_samples = np.linspace(0.4, 0.7)
    qe_curve = np.exp(-((wl_samples - central_frec) ** 2) / (2 * sigma**2))
    img_a = FresnelPropagator(img, ps, wl_samples, z)
    img_t = img_a * np.expand_dims(np.expand_dims(qe_curve, -1), -1)
    img_t = np.nan_to_num(img_t, nan=0.0)
    light_comp: np.ndarray = np.mean(img_t, axis=0)
    return light_comp


def FresnelPropagator(
    E0: np.ndarray, ps: float, lambda0: Union[float, np.ndarray], z: float
) -> np.ndarray:
    # Parameters: E0 - initial complex field in x-y source plane => sqrt(I_0)*e^i phi
    #             ps - pixel size in microns
    #             lambda0 - wavelength in nm
    #             z - z-value (distance from sensor to object)
    #             background - optional background image to divide out from

    k = np.fft.fftfreq(E0.shape[0], ps)
    kxx, kyy = np.meshgrid(k, k)
    e_kxx = cp.asarray(np.expand_dims(kxx, 0))
    e_kyy = cp.asarray(np.expand_dims(kyy, 0))
    e_lam = cp.asarray(np.expand_dims(np.expand_dims(lambda0, -1), -1))
    E0 = cp.asarray(np.expand_dims(E0, 0))
    H = fftshift(
        cp.exp(1j * (2 * np.pi / e_lam) * z)
        * cp.exp(1j * np.pi * e_lam * z * (e_kxx**2 + e_kyy**2)),
        axes=(1, 2),
    )
    E0fft = easy_fft(
        E0, axes=(1, 2)
    )  # Centered about 0 since fx and fy centered about 0
    G = H * E0fft
    Ef = easy_ifft(G, axes=(1, 2))  # Output after deshifting Fourier transform
    I: np.ndarray = (cp.abs(Ef) ** 2).get()
    return I



from typing import Iterator, Tuple

import numpy as np
from skimage.transform import resize

from cvdm.utils.phase_utils import FresnelPropagator, light_component_sim

DZ = -1


class PhasePolychromeDataloader:

    def __init__(
        self,
        path: str,
        n_samples: int,
        im_size: int,
        training: bool,
    ) -> None:

        self._x = np.load(f"{path}/y.npy", mmap_mode="r+")[:n_samples]
        self._n_samples: int = min(n_samples, self._x.shape[0])
        self._training = training
        self._im_size = im_size

    def __len__(self) -> int:
        return self._n_samples

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        x = self._x[idx][:, :, 0]
        x = (x - np.amin(x)) / (np.amax(x) - np.amin(x))
        x = np.nan_to_num(x, nan=0.0)
        x = resize(x, (self._im_size, self._im_size)) * 3.5

        if self._training:
            lam_p = np.random.uniform(0, 0.3)
            dz_r = np.random.uniform(0.01, 3)
            dz_g = np.random.uniform(0.01, 3)
            dz_b = np.random.uniform(0.01, 3)

            cen_r = 0.630
            cen_b = 0.450
            cen_g = 0.550

            rnd_r = np.random.uniform(0.01, 0.1)
            rnd_g = np.random.uniform(0.01, 0.1)
            rnd_b = np.random.uniform(0.01, 0.1)

            img_r = light_component_sim(np.exp(1j * x), 0.21, dz_r, cen_r, rnd_r)

            img_r = img_r + np.random.normal(lam_p, lam_p, size=img_r.shape)
            img_g = light_component_sim(np.exp(1j * x), 0.21, dz_g, cen_g, rnd_g)
            img_g = img_g + np.random.normal(lam_p, lam_p, size=img_g.shape)

            img_b = light_component_sim(np.exp(1j * x), 0.21, dz_b, cen_b, rnd_b)

            img_b = img_b + np.random.normal(lam_p, lam_p, size=img_b.shape)
        else:
            img_r = FresnelPropagator(np.exp(1j * x), 0.21, 0.630, DZ)
            img_g = FresnelPropagator(np.exp(1j * x), 0.21, 0.530, DZ)
            img_b = FresnelPropagator(np.exp(1j * x), 0.21, 0.450, DZ)

        img_stack = np.stack(
            [img_r / np.amax(img_r), img_g / np.amax(img_g), img_b / np.amax(img_b)],
            axis=-1,
        )
        img_stack = np.squeeze(img_stack * 2 - 1)

        y = (x / 3.5) * 2 - 1

        return img_stack, np.expand_dims(y, 2)

    def __call__(self) -> Iterator[Tuple[np.ndarray, np.ndarray]]:
        for i in range(self.__len__()):
            yield self.__getitem__(i)



from typing import Any, Dict, Optional

import yaml

from cvdm.configs.data_config import DataConfig
from cvdm.configs.eval_config import EvalConfig
from cvdm.configs.model_config import ModelConfig
from cvdm.configs.neptune_config import NeptuneConfig
from cvdm.configs.training_config import TrainingConfig


def load_config_from_yaml(yaml_file: str) -> Dict[str, Any]:
    with open(yaml_file, "r") as file:
        config_data: Dict[str, Any] = yaml.safe_load(file)
    return config_data


def create_model_config(config_data: Dict[str, Any]) -> ModelConfig:
    return ModelConfig(**config_data["model"])


def create_training_config(config_data: Dict[str, Any]) -> TrainingConfig:
    return TrainingConfig(**config_data["training"])


def create_data_config(config_data: Dict[str, Any]) -> DataConfig:
    return DataConfig(**config_data["data"])


def create_eval_config(config_data: Dict[str, Any]) -> EvalConfig:
    return EvalConfig(**config_data["eval"])


def create_neptune_config(config_data: Dict[str, Any]) -> Optional[NeptuneConfig]:
    if "neptune" in config_data:
        return NeptuneConfig(**config_data["neptune"])
    else:
        return None



import tensorflow as tf
from tensorflow.keras.layers import Activation, Conv2D, Input
from tensorflow.keras.models import Model


def time_model(
    t_shape: tf.TensorShape, n_hidden: int, act: str = "softplus", mult: bool = True
) -> Model:
    t_input = Input(t_shape)
    l1 = Conv2D(
        1,
        kernel_size=(1, 1),
        padding="same",
        kernel_constraint=tf.keras.constraints.NonNeg(),
    )(t_input)
    x = Conv2D(
        n_hidden,
        kernel_size=(3, 3),
        padding="same",
        activation="sigmoid",
        kernel_constraint=tf.keras.constraints.NonNeg(),
    )(l1)
    x = Conv2D(
        1,
        kernel_size=(1, 1),
        padding="same",
        kernel_constraint=tf.keras.constraints.NonNeg(),
    )(x)
    if mult:
        out = Activation(act)(x + l1) * t_input
    else:
        out = Activation(act)(x + l1)
    model: Model = Model(t_input, out)
    return model



from typing import Optional

import tensorflow as tf
from tensorflow.keras.activations import swish
from tensorflow.keras.layers import Concatenate, Conv2D, Dropout
from tensorflow_addons.layers import GroupNormalization

from cvdm.architectures.components.conditional_instance_normalization import (
    ConditionalInstanceNormalization,
)


def deep_resblock(
    x: tf.Tensor, noise_embedding: tf.Tensor, n_out_channels: int, dropout: float = 0.0
) -> tf.Tensor:
    """
    Reuse the resblock definitions from https://arxiv.org/pdf/1809.11096.pdf
    Slight tweak to use ConditionalInstanceNormalization and pre-activation
    as described in https://arxiv.org/pdf/1907.05600.pdf
    """
    assert n_out_channels is not None
    track_a = Conv2D(n_out_channels, 1, padding="same")(x)

    track_b = GroupNormalization(groups=32, axis=-1)(x)
    track_b = swish(track_b)
    track_b = Conv2D(n_out_channels, 1, padding="same")(track_b)
    track_b = ConditionalInstanceNormalization()([track_b, noise_embedding])
    track_b = swish(track_b)
    track_b = Conv2D(n_out_channels, 3, padding="same")(track_b)
    track_b = Dropout(dropout)(track_b)
    track_b = GroupNormalization(groups=32, axis=-1)(x)
    track_b = swish(track_b)
    track_b = Conv2D(n_out_channels, 3, padding="same")(track_b)
    track_b = ConditionalInstanceNormalization()([track_b, noise_embedding])
    track_b = swish(track_b)
    track_b = Conv2D(n_out_channels, 1, padding="same")(track_b)
    out: tf.Tensor = track_a + track_b
    return out


def up_deep_resblock(
    x: tf.Tensor,
    skip_x: tf.Tensor,
    noise_embedding: tf.Tensor,
    n_out_channels: int,
    dropout: float = 0.0,
) -> tf.Tensor:

    skip_x = skip_x / tf.math.sqrt(2.0)
    x = Concatenate(axis=-1)([x, skip_x])

    return deep_resblock(
        x, noise_embedding, n_out_channels=n_out_channels, dropout=dropout
    )



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow.keras.layers import Input, Lambda
from tensorflow.keras.models import Model

from cvdm.configs.model_config import ModelConfig
from cvdm.diffusion_models.mean_model import mean_model
from cvdm.diffusion_models.noise_model import noise_model
from cvdm.diffusion_models.variance_model import variance_model
from cvdm.utils.data_utils import obtain_noisy_sample
from cvdm.utils.loss_utils import linear_loss, normal_kl
from cvdm.utils.training_utils import taylor_expand_gamma, time_grad


def create_joint_model(
    input_shape_condition: tf.TensorShape,
    timesteps: int,
    out_channels: int,
    model_config: ModelConfig,
) -> Tuple[Model, Model, Model, Optional[Model]]:
    """This function creates a Keras model for the image denoising pipeline."""

    ground_truth = Input(input_shape_condition[:-1] + (out_channels,))
    dirty_img = Input(input_shape_condition)

    timesteps_Lt = Input(input_shape_condition[:-1] + (out_channels,))
    timesteps_LT = tf.ones_like(timesteps_Lt) - (1 / timesteps)
    timesteps_L0 = tf.zeros_like(timesteps_Lt) + (1 / timesteps)

    sch_model = variance_model(input_shape_condition, out_channels)
    sch_params_LT = sch_model([dirty_img, timesteps_LT])
    sch_params_Lt = sch_model([dirty_img, timesteps_Lt])
    sch_params_L0 = sch_model([dirty_img, timesteps_L0])

    if model_config.zmd:
        pass

        sigma = 0.5
        n_model = noise_model(
            input_shape_condition,
            out_channels,
            model_type=model_config.noise_model_type,
            zmd=model_config.zmd,
        )
        mu_model = mean_model(input_shape_condition, out_channels)
        mean_pred = mu_model(dirty_img)[0]
        mean_pred_sg = tf.stop_gradient(mean_pred)
        n_sample_LT = Lambda(obtain_noisy_sample)(
            [(ground_truth - mean_pred_sg) / sigma, sch_params_LT[0]]
        )
        n_sample_Lt = Lambda(obtain_noisy_sample)(
            [(ground_truth - mean_pred_sg) / sigma, sch_params_Lt[0]]
        )
        pred_noise_Lt = n_model(
            [n_sample_Lt[0], dirty_img, mean_pred_sg, sch_params_Lt[0]]
        )

    else:
        n_model = noise_model(
            input_shape_condition,
            out_channels,
            model_type=model_config.noise_model_type,
        )

        n_sample_LT = Lambda(obtain_noisy_sample)([ground_truth, sch_params_LT[0]])
        n_sample_Lt = Lambda(obtain_noisy_sample)([ground_truth, sch_params_Lt[0]])
        pred_noise_Lt = n_model([n_sample_Lt[0], dirty_img, sch_params_Lt[0]])

    d_alpha_t = Lambda(time_grad)([sch_params_Lt[0], timesteps_Lt])

    gamma_exp = Lambda(
        taylor_expand_gamma, arguments={"n": model_config.snr_expansion_n}
    )(sch_params_Lt[0])
    d_alpha_n_t = Lambda(time_grad)([gamma_exp, timesteps_Lt])
    d2_alpha_n_t = Lambda(time_grad)([d_alpha_n_t, timesteps_Lt])

    kl_divergence_T = Lambda(normal_kl)(
        [
            tf.zeros_like(n_sample_LT[2]),
            tf.zeros_like(n_sample_LT[2]),
            n_sample_LT[2],
            tf.math.log(n_sample_LT[3] + 1e-7),
        ]
    )

    delta_noise = tf.square(pred_noise_Lt - n_sample_Lt[1])
    L_beta = (
        tf.square(d_alpha_t + sch_params_Lt[1] * sch_params_Lt[0])
        + tf.square(sch_params_L0[0] - 1)
        + tf.square(sch_params_LT[0])
    )
    L_gamma = model_config.alpha * tf.square(d2_alpha_n_t)
    joint_model: Model
    if model_config.zmd:
        delta_mean = tf.square(ground_truth - mean_pred)
        joint_model = Model(
            [ground_truth, dirty_img, timesteps_Lt],
            [delta_noise, L_beta, kl_divergence_T, L_gamma, delta_mean],
        )
        return n_model, joint_model, sch_model, mu_model
    else:
        joint_model = Model(
            [ground_truth, dirty_img, timesteps_Lt],
            [delta_noise, L_beta, kl_divergence_T, L_gamma],
        )
        joint_model.summary()
        return n_model, joint_model, sch_model, None


def instantiate_cvdm(
    lr: float,
    generation_timesteps: int,
    cond_shape: tf.TensorShape,
    out_shape: tf.TensorShape,
    model_config: ModelConfig,
) -> Tuple[Model, Model, Model, Optional[Model]]:
    opt_m = tf.keras.optimizers.Adam(learning_rate=lr)
    out_channels = out_shape[-1]
    assert out_channels is not None

    if model_config.zmd:
        models = create_joint_model(
            cond_shape, generation_timesteps, out_channels, model_config
        )
        noise_model, joint_model, schedule_model, mu_model = models
        joint_model.compile(loss=linear_loss, loss_weights=[1, 2, 2, 2, 2], optimizer=opt_m)  # type: ignore
    else:
        models = create_joint_model(
            cond_shape, generation_timesteps, out_channels, model_config
        )
        noise_model, joint_model, schedule_model, _ = models
        joint_model.compile(loss=linear_loss, loss_weights=[1, 2, 2, 2], optimizer=opt_m)  # type: ignore

    return models



from typing import Iterator, Tuple

import numpy as np
from PIL import Image

from cvdm.utils.data_utils import read_and_patch_image_from_filename


class ImageDirDataloader:
    def __init__(self, paths: np.ndarray, n_samples: int, im_size: int) -> None:
        self._n_samples = min(n_samples, len(paths))
        self._paths = paths[:n_samples]
        self._im_size = im_size

    def __len__(self) -> int:
        return self._n_samples

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        filename = self._paths[idx]
        img_patch = read_and_patch_image_from_filename(filename, self._im_size)

        patch_LR = img_patch.resize(
            (self._im_size // 4, self._im_size // 4), Image.LANCZOS
        )
        err_HR = patch_LR.resize((self._im_size, self._im_size), Image.LANCZOS)
        img_patch_arr = np.array(img_patch)
        err_HR_arr = np.array(err_HR)
        return (err_HR_arr / 255.0) * 2 - 1, (img_patch_arr / 255.0) * 2 - 1

    def __call__(self) -> Iterator[Tuple[np.ndarray, np.ndarray]]:
        for i in range(self.__len__()):
            yield self.__getitem__(i)



from typing import Iterator, Tuple

import cv2
import numpy as np

from cvdm.utils.data_utils import read_and_patch_image_from_filename
from cvdm.utils.phase_utils import FresnelPropagator

WL = 0.521
DZ = -1


class Phase2ShotDataloader:

    def __init__(self, paths: np.ndarray, n_samples: int, im_size: int, training: bool):
        self._n_samples = min(n_samples, len(paths))
        self._paths = paths[:n_samples]
        self._im_size = im_size
        self._training = training

    def __len__(self) -> int:
        return self._n_samples

    def __getitem__(self, idx: int) -> Tuple[np.ndarray, np.ndarray]:
        filename = self._paths[idx]
        img_patch = read_and_patch_image_from_filename(filename, self._im_size)

        img_patch_arr: np.ndarray = np.array(img_patch)

        img_patch_arr = cv2.cvtColor(img_patch_arr, cv2.COLOR_RGB2GRAY)
        x = np.divide(img_patch_arr, 255.0)
        x = (x - np.amin(x)) / (np.amax(x) - np.amin(x))
        x = np.nan_to_num(x, nan=0.0) * 3.5

        if self._training:
            lam = np.random.uniform(0, 0.2)
            img_f = FresnelPropagator(np.exp(1j * x), 0.21, np.asarray([WL]), DZ)[
                0
            ] + np.random.normal(lam, lam, x.shape)
            img_b = FresnelPropagator(np.exp(1j * x), 0.21, np.asarray([WL]), -DZ)[
                0
            ] + np.random.normal(lam, lam, x.shape)
        else:
            img_f = FresnelPropagator(np.exp(1j * x), 0.21, np.asarray([WL]), DZ)[0]
            img_b = FresnelPropagator(np.exp(1j * x), 0.21, np.asarray([WL]), -DZ)[0]

        img_stack = np.stack((img_f, img_b), axis=-1)
        img_stack = (img_stack - np.amin(img_stack)) / (
            np.amax(img_stack) - np.amin(img_stack)
        )
        img_stack = np.squeeze(img_stack * 2 - 1)
        y = (x / 3.5) * 2 - 1

        return img_stack, np.expand_dims(y, 2)

    def __call__(self) -> Iterator[Tuple[np.ndarray, np.ndarray]]:

        for i in range(self.__len__()):
            yield self.__getitem__(i)
